/* Msg, a simple program for sending messages to other users
 * (c) Juho Snellman (jsnell@lyseo.edu.ouka.fi) 1998. */

/* For licensing information, see the file COPYING */

#include <getopt.h>
#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <utmp.h>
#include <readline/readline.h>

/* Should be defined in stdio.h, but isn't. */
char *cuserid (char *);

#define MAXTTYS 256   /* Is there any real definition for this ? */

typedef struct
{
    char* name;       /* name (no path) of the tty */
    long int idle;    /* seconds since last access to tty */
}
tty_info;

int tty_select (tty_info**);
int tty_info_asc (tty_info**, tty_info**);
int tty_info_desc (tty_info**, tty_info**);
int tty_path (char*, char*, int);
int tty_uname_writable(char*, tty_info**, int);
int tty_uname(char*, tty_info**);
FILE* write_message (void);
FILE* open_lastmsg(char*);
int send_message (tty_info**, FILE*);
void print_usage(void);
void drop_root(void);
void get_root(void);

uid_t ugid;
uid_t egid;

int main (int argc, char** argv)
{
    tty_info *tty_list[MAXTTYS];
    char *uname = NULL;
    tty_info *t;
    int i, send_old = 0, use_tty = 0, c;
    FILE *msgfile;
    
    ugid = getgid();
    egid = getegid();
    
    drop_root();
    
    for (i = 0; i < MAXTTYS; i++ )
    {
	tty_list[i] = NULL;
    }
    
    if (argc < 2)
    {
	print_usage();
	return -1;
    }
    
    i = 0;
    while (1)
    {
	c = getopt(argc, argv, "rt:u:");
	if (c == -1)
	  break;
	
	switch (c)
	{
	 case 'h':
	    print_usage();
	    return -1;
	    break;
	 case 'r':
	    send_old = 1;
	    break;
	 case 't':
	    i++;
	    t = malloc(sizeof(tty_info) );
	    if (t == NULL)
	    {
		fprintf(stderr, "Out of memory ?");
		return -1;
	    }			
	    t->name = malloc(strlen(optarg)+1);
	    if (t->name == NULL)
	    {
		fprintf(stderr, "Out of memory ?");
		return -1;
	    }			
	    strcpy(t->name, optarg);
	    t->idle = 0;
	    tty_list[0] = t;
	    use_tty = 1;
	    break;
	 case 'u':
	    i++;
	    if (!uname)
	    {
		uname = malloc(strlen(optarg)+1 );
		if (uname == NULL)
		{
		    fprintf(stderr, "Out of memory ?\n");
		    return -1;
		}
		strcpy(uname, optarg);
	    }
	    break;
	}
	i++;
    }
    
    if (argc == i+1)
    {
	print_usage();
	return -1;
    }
    
    if (!uname)
    {
	uname = malloc(strlen(argv[i+1])+1 );
	if (uname == NULL)
	{
	    fprintf(stderr, "Out of memory ?\n");
	    return -1;
	}
	strcpy(uname, argv[i+1]);
    }
    
    
    if ( getpwnam(uname) == NULL)
    {
	fprintf(stderr, "Unknown user '%s'.\n", uname);
	return -1;
    }
    
    if ( !use_tty )
    {
	if ( !tty_uname_writable (uname, tty_list, 1) )
	  return -1;
    }
    else
    {  
	if ( !tty_uname_writable (uname, tty_list, 0) )
	  return -1;
    }
    
    if (!send_old)
      msgfile = write_message();
    else
      msgfile = open_lastmsg("r");
    
    if (msgfile == NULL)
      return -1;
    
    if ( !use_tty )
    {
	if ( !tty_uname_writable (uname, tty_list, 1) )
	  return -1;
    }
    else
    {  
	if ( !tty_uname_writable (uname, tty_list, 0) )
	  return -1;
    }
    
    tty_select(tty_list);
    
    send_message(tty_list, msgfile);
    
    return 0;
}

void drop_root (void)
{    
    /* printf("Changing to gid %d\n", ugid); */

    if (setegid(ugid) < 0)
    {
	fprintf(stderr, "Error dropping tty priviledges: %s\n", 
		strerror(errno));
	exit(EXIT_FAILURE);
    }        
}

void get_root (void)
{
    /* printf("Returning to egid %d\n", egid); */

    if (setegid(egid) < 0)
    {
	fprintf(stderr, "Error restoring root priviledges: %s\n", 
		strerror(errno));
	exit(EXIT_FAILURE);
    }        
}

/* Return tty with _greatest_ idle time (< 5 min). I hate it when people 
 * clobber my most used tty:s :-) [Should probably insert some 
 * extra checks here for usability] */
int tty_select (tty_info **tty_list)
{
    int i, j;
    char path[256];
    struct stat tty_st;
    
    for (i = 0; tty_list[i] != NULL; i++)
    {
	if ( tty_path(tty_list[i]->name, path, sizeof(path)) < 0)
	{
	    fprintf(stderr, "Too long tty-name '%s'.\n", tty_list[i]->name);
	    return -1;
	}
	if ( stat(path, &tty_st) == 0)
	{
	    tty_list[i]->idle = time(NULL)-tty_st.st_atime;
	}
    }
    
    qsort(tty_list, i, sizeof(tty_info*), tty_info_asc);
    
    for (j = 0; j < i; j++)
      if (tty_list[j]->idle > 5*60)
	break;
    
    if (j)
      qsort(tty_list, j, sizeof(tty_info*), tty_info_desc);
    if (i-j)
      qsort(tty_list+sizeof(tty_info*)*j, i-j, sizeof(tty_info*), tty_info_asc);
    
    return 0;
}

/* Comparison function for tty_select(). */
int tty_info_desc (tty_info **a, tty_info **b)
{
    /* Sorts any null-entries last. Maybe :) */
    if ( a == NULL && b == NULL)
      return 0;
    if ( *a == NULL && *b == NULL)
      return 0;
    if ( *a == NULL)
      return -1;
    if ( *b == NULL)
      return 1;
    
    return (*b)->idle - (*a)->idle;
}

/* Comparison function for tty_select(). */
int tty_info_asc (tty_info **a, tty_info **b)
{	
    /* Sorts any null-entries last. Maybe :) */
    if ( a == NULL && b == NULL)
      return 0;
    if ( *a == NULL && *b == NULL)
      return 0;
    if ( *a == NULL)
      return 1;
    if ( *b == NULL)
      return -1;
    
    return (*a)->idle - (*b)->idle;
}

/* Prepend tty-name with /dev/ */
int tty_path (char *tty, char *s, int len)
{
    if (len < strlen("/dev/") + strlen(tty) + 1)
    {
	s[0] = '\0';
	return -1;
    }
    sprintf(s, "/dev/%s", tty);
    return 0;
}

/* Removes all unwritable tty:s from tty_list */
int tty_uname_writable(char *uname, tty_info **tty_list, int rebuild)
{
    int i, amount;
    
    if (rebuild) /* Get new list of tty:s */
      amount = tty_uname(uname, tty_list);
    else
      for (amount = 0; tty_list[amount] != NULL && amount < MAXTTYS; amount++)
	;
    
    if (amount == 0)
    {
	fprintf(stderr, "User '%s' not logged in.\n", uname);
	return 0;
    }
    
    for (i = 0; i < amount && tty_list[i] != NULL; i++)
    {
	char path[256];
	struct stat s;
	
	if ( tty_path(tty_list[i]->name, path, sizeof(path)) < 0)
	{
	    fprintf(stderr, "Too long tty-name '%s'.\n", tty_list[i]->name);
	    return 0;
	}
		
	if ( stat(path, &s) != 0 || !(s.st_mode & S_IWGRP))
	{
	    tty_list[i] = tty_list[amount];
	    tty_list[amount] = NULL;
            amount--;
	    i--;
	}
    }
    
    if (amount == 0)
    {
	fprintf(stderr, "User '%s' has messages disabled.\n", uname);
	return 0;
    }
    
    return 1;
}

/* Enter all tty:s used by user into tty_list */
int tty_uname(char *uname, tty_info **tty_list)
{
    struct utmp *u;
    tty_info *t;
    int i = 0;
    
    setutent();
    
    while( (u = getutent()) != NULL && i < MAXTTYS)
    {	  
	if (strcmp (uname, u->ut_user) == 0)
	{
	    t = malloc(sizeof(tty_info) );
	    if (t == NULL)
	    {
		fprintf(stderr, "Out of memory ?");
		return -1;
	    }			
	    t->name = malloc(strlen(u->ut_line)+1);
	    if (t->name == NULL)
	    {
		fprintf(stderr, "Out of memory ?");
		return -1;
	    }			
	    strcpy(t->name, u->ut_line);
	    t->idle = 0;
	    tty_list[i++] = t;
	}		
    }
    
    endutent();
    return i;
}

/* Read message from user (with readline), write it to ~/.lastmsg
 * until EOF or an empty line, and return a pointer to the file 
 * re-opened as read-only */
FILE* write_message(void)
{
    FILE *msgfile; 
    char *s;
    
    msgfile = open_lastmsg("w");
    
    rl_bind_key ('\t', rl_insert);
    rl_variable_bind("convert-meta", "off");
    rl_variable_bind("output-meta", "on");
    rl_variable_bind("input-meta", "on"); 
    
    while ( (s = readline("msg > ") ) != NULL && strlen(s) > 0)
    {
	fprintf(msgfile, "%s\r\n", s); /* '\r' for non-sane ttys */
	free(s);
    }
    
    fclose(msgfile);
    
    msgfile = open_lastmsg("r");
    
    return msgfile;
}

/* Open $HOME/.lastmsg */
FILE* open_lastmsg(char* perm)
{
    FILE* msgfile;
    struct passwd *pw;
    char *path;
        
    if ( ( pw = getpwnam(cuserid(NULL) )) == NULL)
    {
	fprintf(stderr, "Unknown user '%s'.\n", cuserid(NULL) );
    }
    
    path = malloc( strlen(pw->pw_dir) + strlen("/.lastmsg") + 1);
    if (path == NULL)
    {
	fprintf(stderr, "Out of memory ?");
	return NULL;
    }			
    
    sprintf(path, "%s/%s", pw->pw_dir, ".lastmsg");
    
    msgfile = fopen(path, perm);
    
    if (msgfile == NULL)
    {
	fprintf(stderr, "Error opening '%s': %s\n", path, 
		sys_errlist[errno]);
	return NULL;
    }			
        
    return msgfile;
}


/* Go through the tty_list, try to write message to each one. Return
 * when successfull */
int send_message(tty_info **tty_list, FILE* msgfile)
{
    FILE *ttyfile; 
    char path[256];
    char s[256];
    char *uname, *ttyn;
    int i;
    time_t t1;
    struct tm *t2;
    
    uname = cuserid(NULL);
    if (uname == NULL)
      uname = "____";
    
    ttyn = ttyname(0);
    if (ttyn == NULL)
      ttyn = ttyname(1);
    if (ttyn == NULL)
      ttyn = ttyname(2);
    if (ttyn == NULL)
      ttyn = "tty";
    else
      ttyn += strlen("/dev/");
    
    for (i = 0; tty_list[i] != NULL; i++)
    {
	if ( tty_path(tty_list[i]->name, path, sizeof(path)) == 0)
	{
	    get_root();
	    ttyfile = fopen(path, "w");
	    drop_root();
	    
	    if (ttyfile != NULL)
	    {
		t1 = time(NULL);
		t2 = localtime( &t1 );
		fprintf(ttyfile, "\r\n\07Message from %s@%s (%02d:%02d) :\r\n",
			uname, ttyn, t2->tm_hour, t2->tm_min);
		while ( fgets(s, sizeof(s), msgfile) )
		  fprintf(ttyfile, s);
		fprintf(ttyfile, "\n");
		fclose(ttyfile);
		return 0;
	    }				
	}
    }
    
    fprintf(stderr, "Message not sent: %s\n", sys_errlist[errno]);
    
    return 0;
}

void print_usage(void)
{
    printf("msg usage: msg [options] uname.\n");
    printf("Sends a message to another user's tty.\n");
    printf("\n");
    printf("-h,\t\tshow this message\n");
    printf("-r,\t\tsend message from .lastmsg\n");
    printf("-t [tty],\tsend message to a specific tty\n");
    printf("-u [uname],\tlonger method for specifying the username\n");
    printf("\n");
}

